-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- 
-- $Id: long_xact.sql.in.c 9324 2012-02-27 22:08:12Z pramsey $
--
-- PostGIS - Spatial Types for PostgreSQL
-- http://postgis.refractions.net
-- Copyright 2001-2003 Refractions Research Inc.
--
-- This is free software; you can redistribute and/or modify it under
-- the terms of the GNU General Public Licence. See the COPYING file.
--  
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

#include "sqldefines.h"

-----------------------------------------------------------------------
-- 
-- LONG TERM LOCKING
-- 
-- This module and associated pl/pgsql functions have been implemented
-- to provide long locking support required by Web Feature Service 
-- specification
-- 
-- See: https://portal.opengeospatial.org/files/?artifact_id=7176
-- 
-- It is based on original work by David Blasby <dblasby@openplans.org>
-- and has been modified and packaged by Sandro Santilli <strk@refractions.net>
-- and Carlos Ruiz <cruizch@gmail.com>
--
-- WARNING! users must use serializable transaction level otherwise locking 
-- mechanism would break
-- 
-- See: http://www.postgresql.org/docs/7.4/static/transaction-iso.html
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- postgis_UnlockRows(<authid>)
-- removes all locks held by the given auth
-- returns the number of locks released
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_UnlockRows(TEXT)
	RETURNS INTEGER
	AS $$ 
DECLARE
	ret INTEGER;
BEGIN

	IF NOT postgis_LongTransactionsEnabled() THEN
		RAISE EXCEPTION 'Long transaction support disabled, use postgis_EnableLongTransaction() to enable.';
	END IF;

	EXECUTE 'DELETE FROM authorization_table where authid = ' ||
		quote_literal($1);

	GET DIAGNOSTICS ret = ROW_COUNT;

	RETURN ret;

END;
$$
LANGUAGE 'plpgsql'  VOLATILE STRICT;

-----------------------------------------------------------------------
-- postgis_LockRow(<schema>, <table>, <rowid>, <auth>, <expires>)
-- set lock/authorization for specific row in table
-- returns 1 if successfully obtained the lock, 0 otherwise
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_LockRow(TEXT, TEXT, TEXT, TEXT, TIMESTAMP)
	RETURNS INTEGER
	AS $$ 
DECLARE
	myschema ALIAS FOR $1;
	mytable ALIAS FOR $2;
	myrid ALIAS FOR $3;
	authid ALIAS FOR $4;
	expires ALIAS FOR $5;
	mytoid OID;
	tauthid TEXT;
	ret INTEGER;
BEGIN

	IF NOT postgis_LongTransactionsEnabled() THEN
		RAISE EXCEPTION 'Long transaction support disabled, use postgis_EnableLongTransaction() to enable.';
	END IF;
	
	EXECUTE 'DELETE FROM authorization_table WHERE expires < now()'; 
	
	SELECT pg_class.oid INTO mytoid FROM pg_class, pg_namespace 
		WHERE pg_class.relname = mytable 
		AND pg_class.relnamespace = pg_namespace.oid 
		AND pg_namespace.nspname = myschema;
	
	SELECT authid INTO tauthid FROM authorization_table WHERE toid = mytoid AND rid = myrid;
	
	IF tauthid IS NOT NULL THEN 
		IF authid == tauthid THEN 
			RETURN 0;
		ELSE 
			RETURN 1;
		END IF;
	END IF;
	
	EXECUTE 'INSERT INTO authorization_table VALUES (' ||
		quote_literal(mytoid::text) || ',' || quote_literal(myrid) || ',' ||
		quote_literal(expires::text) || ',' ||
		quote_literal(authid) || ')';
	
	GET DIAGNOSTICS ret = ROW_COUNT;

	RETURN ret;

END;
$$
LANGUAGE 'plpgsql'  VOLATILE STRICT;

-----------------------------------------------------------------------
-- postgis_LockRow(<schema>, <table>, <rid>, <authid>)
-- set lock/authorization for specific row in table
-- returns 1 if successfully obtained the lock, 0 otherwise
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_LockRow(TEXT, TEXT, TEXT, TEXT)
	RETURNS INTEGER
	AS
$$ SELECT postgis_LockRow($1, $2, $3, $4, now()::timestamp + '1:00'); $$
	LANGUAGE 'sql'  VOLATILE STRICT;

-----------------------------------------------------------------------
-- postgis_LockRow(<table>, <rid>, <authid>)
-- set lock/authorization for specific row in table
-- returns 1 if successfully obtained the lock, 0 otherwise
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_LockRow(TEXT, TEXT, TEXT)
	RETURNS INTEGER
	AS
$$ SELECT postgis_LockRow(current_schema(), $1, $2, $3, now()::timestamp + '1:00'); $$
	LANGUAGE 'sql'  VOLATILE STRICT;

-----------------------------------------------------------------------
-- postgis_LockRow(<table>, <rid>, <authid>, <expires>)
-- set lock/authorization for specific row in table
-- returns 1 if successfully obtained the lock, 0 otherwise
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_LockRow(TEXT, TEXT, TEXT, TIMESTAMP)
	RETURNS INTEGER
	AS
$$ SELECT postgis_LockRow(current_schema(), $1, $2, $3, $4); $$
	LANGUAGE 'sql'  VOLATILE STRICT;

-----------------------------------------------------------------------
-- postgis_AddAuth(<auth>)
-- adds an authorization token to be used in current transaction
-- always returns true
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_AddAuth(TEXT)
	RETURNS BOOLEAN
	AS $$ 
DECLARE
	lockid ALIAS FOR $1;
BEGIN

	IF NOT EXISTS(SELECT * FROM pg_class WHERE relname = 'temp_lock_have_table') THEN 
		EXECUTE 'CREATE TEMP TABLE temp_lock_have_table (' || 
			' transid XID NOT NULL, lockcode TEXT NOT NULL' || 
			' )' || 
			' ON COMMIT DELETE ROWS'; -- this works just for PostgreSQL 7.4 up
	END IF;
	
	INSERT INTO temp_lock_have_table VALUES (getTransactionID(), lockid);
	
	RETURN TRUE;

END;
$$
LANGUAGE PLPGSQL;

-----------------------------------------------------------------------
-- postgis_CheckAuth(<schema>, <table>, <ridcolumn>)
-- check updates and deletes of rows in given table for being authorized
-- always returns 0
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_CheckAuth(TEXT, TEXT, TEXT)
	RETURNS INT
	AS $$ 
DECLARE
	schema TEXT;
	exist BOOLEAN;
BEGIN

	IF NOT postgis_LongTransactionsEnabled() THEN
		RAISE EXCEPTION 'Long transaction support disabled, use postgis_EnableLongTransaction() to enable.';
	END IF;

	if ($1 != '') THEN
		schema = $1;
	ELSE
		SELECT current_schema() into schema;
	END IF;
	
	EXECUTE 'SELECT EXISTS(SELECT * FROM pg_namespace, pg_class, pg_trigger WHERE' || 
		' pg_namespace.nspname = ' || quote_literal(schema) || 
		' AND pg_class.relname = ' || quote_literal($2) || 
		' AND pg_class.oid = pg_trigger.tgrelid' || 
		' AND pg_trigger.tgname = ' || quote_literal('check_auth') || ')' INTO exist;
	
	IF NOT exist THEN
		EXECUTE 'CREATE TRIGGER check_auth BEFORE UPDATE OR DELETE ON' || 
			' ' || quote_ident(schema) || '.' || quote_ident($2) || 
			' FOR EACH ROW EXECUTE PROCEDURE CheckAuthTrigger(' || quote_literal($3) || ')';
	END IF;
	
	RETURN 0;

END;
$$
LANGUAGE 'plpgsql';

-----------------------------------------------------------------------
-- postgis_CheckAuth(<table>, <ridcolumn>)
-- create a trigger to check if a row can be updated or deleted
-- always returns 0
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_CheckAuth(text, text)
	RETURNS INTEGER
	AS
	$$ SELECT postgis_CheckAuth('', $1, $2) $$
	LANGUAGE 'SQL';

-----------------------------------------------------------------------
-- CheckAuthTrigger()
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION CheckAuthTrigger()
	RETURNS TRIGGER AS 
	'MODULE_PATHNAME', 'check_authorization'
	LANGUAGE C;

-----------------------------------------------------------------------
-- GetTransactionID()
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION GetTransactionID()
	RETURNS XID AS 
	'MODULE_PATHNAME', 'getTransactionID'
	LANGUAGE C;

-----------------------------------------------------------------------
-- enable long transactions support
-- creates the authorization_table if not exists
-- returns 'Long transactions support enabled'
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_EnableLongTransactions()
	RETURNS TEXT
	AS $$ 
BEGIN
	
	IF NOT EXISTS(SELECT * FROM pg_class WHERE relname = 'authorization_table') THEN 
		EXECUTE 'CREATE TABLE authorization_table (
			toid OID NOT NULL, 
			rid TEXT NOT NULL, 
			expires TIMESTAMP NOT NULL, 
			authid TEXT NOT NULL
		)';
	END IF;
	
	IF NOT EXISTS(SELECT * FROM pg_class WHERE relname = 'authorized_tables') THEN 
		EXECUTE 'CREATE VIEW authorized_tables AS' || 
			' SELECT' || 
			' pg_namespace.nspname AS schema,' || 
			' pg_class.relname AS table,' || 
			' trim(' || quote_literal(chr(92) || '000') || ' from pg_trigger.tgargs) AS id_column' || 
			' FROM pg_trigger, pg_class, pg_proc, pg_namespace ' || 
			' WHERE pg_proc.proname = ' || quote_literal('checkauthtrigger') || 
			' AND pg_class.relnamespace = pg_namespace.oid' || 
			' AND pg_trigger.tgfoid = pg_proc.oid' || 
			' AND pg_trigger.tgrelid = pg_class.oid';
	END IF;
	
	RETURN 'Long transactions support enabled';

END;
$$
LANGUAGE 'plpgsql';

-----------------------------------------------------------------------
-- postgis_LongTransactionsEnabled()
-- check if long transactions support is enabled
-- returns true if the authorized_tables table exists, false otherwise
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_LongTransactionsEnabled()
	RETURNS BOOLEAN
AS $$ 
BEGIN

	RETURN EXISTS(SELECT * FROM pg_class WHERE relname = 'authorized_tables');

END;
$$
LANGUAGE 'plpgsql';

-----------------------------------------------------------------------
-- disable long transactions support
-- drops the created triggers, the authorization_table table and the authorized_tables view
-- returns 'Long transactions support disabled' on success
-----------------------------------------------------------------------
CREATE OR REPLACE FUNCTION postgis_DisableLongTransactions()
	RETURNS TEXT
	AS $$ 
DECLARE
	rec RECORD;
BEGIN

	FOR rec IN
		SELECT pg_class.relname, pg_trigger.tgname, pg_trigger.tgargs 
			FROM pg_trigger, pg_class, pg_proc 
			WHERE pg_proc.proname = 'checkauthtrigger' 
			AND pg_trigger.tgfoid = pg_proc.oid 
			AND pg_trigger.tgrelid = pg_class.oid
	LOOP
		EXECUTE 'DROP TRIGGER ' || quote_ident(rec.tgname) || 
			' ON ' || quote_ident(rec.relname);
	END LOOP;
	
	DROP TABLE IF EXISTS authorization_table;
	DROP VIEW IF EXISTS authorized_tables;
	
	RETURN 'Long transactions support disabled';

END;
$$
LANGUAGE 'plpgsql';

---------------------------------------------------------------
-- END
---------------------------------------------------------------
